#include <string.h>
#include <unistd.h>
#include <signal.h>
#include "osipua.h"
#include "sdphandler.h"
#include <osip/port.h>

void server(void);
void client(char *url);
void stop_func(int signum);
int invite_cb(OsipDialog *cl,transaction_t *trn, sip_t *msg,void*p);
int bye_cb(OsipDialog *cl,transaction_t *trn, sip_t *msg,void*p);
int invite_accepted_cb(OsipDialog *cl,transaction_t *trn, sip_t *msg,void*p);
int run_cond=1;


static char *usage="usage: osipua_tester [-c] [-s] <sip url>\n";

char *local_sipurl;

int main(int argc, char *argv[])
{
	url_t *url;
	int err;
	char *loc,*rem;
	
	url_init(&url);
	osipua_init();
	signal(SIGINT,stop_func);
	if (argc<2)
	{
		printf(usage);
		return(-1);
	}
	
	/* get local address */
	loc=getenv("SIPURL");
	if (loc!=NULL){
		err=url_parse(url,loc);
		if (err<0){
			printf("Env variable SIPURL is set to a bad formuled sip url: %s\n",loc);
			return -1;
		}
		url_free(url);
		sfree(url);
		local_sipurl=sgetcopy(loc);
	}
	else
	{
		printf("Env variable SIPURL is not set. Using <sip:toto@127.0.0.1> as a default local sip url.\n");
		local_sipurl=sgetcopy("<sip:toto@127.0.0.1>");
	}
	TRACE_INITIALIZE(OSIP_INFO4,NULL);
	if (strcmp(argv[1],"-s")==0) server();
	else if (strcmp(argv[1],"-c")==0) {
		if (argc<3) {
			printf(usage);
			return -1;
		}
		rem=sgetcopy(argv[2]);
		url_init(&url);
		err=url_parse(url,rem);
		if (err<0){
			printf("Bad sip url: %s\n",argv[2]);
			return -1;
		}
		url_free(url);
		client(rem);
	}
	else return(-1);
	return(0);
}


sdp_t * generate_client_offer(SdpContext *ctx, sdp_t *template)
{
	/* the sdp handler propose us a template of sdp. If we use it, we just have to add
	the media streams. If not create a new sdp.*/
	/* we want an audio stream: */
	sdp_m_media_add(template, sgetcopy("audio"),
		       sgetcopy("8002"), NULL, sgetcopy("RTP/AVP"));
	/* with codecs 0,3,8 and lpc10-15 */
	sdp_m_payload_add(template,0,sgetcopy("0"));
	sdp_m_payload_add(template,0,sgetcopy("3"));
	sdp_m_payload_add(template,0,sgetcopy("8"));
	sdp_m_payload_add(template,0,sgetcopy("115"));
	
	/* add some attributes for better understanding with the remote server*/
	sdp_a_attribute_add(template,0, sgetcopy("rtpmap"), sgetcopy("0 pcmu/8000/1"));
	sdp_a_attribute_add(template,0, sgetcopy("rtpmap"), sgetcopy("3 gsm/8000/1"));
	sdp_a_attribute_add(template,0, sgetcopy("rtpmap"), sgetcopy("8 pcma/8000/1"));
	sdp_a_attribute_add(template,0, sgetcopy("rtpmap"), sgetcopy("115 lpc10-15/8000/1"));
	return template;
}

void client(char *callee)
{
	OsipUA *ua;
	OsipDialog *dialog;
	BodyHandler *sdph;

	/* create the ua */
	ua=osip_ua_new();
	osip_ua_set_contact(ua,local_sipurl);
	osip_ua_set_max_dialogs(ua,1);
	/* set some callbacks */
	osip_ua_signal_connect(ua,"INVITE_ACCEPTED",invite_accepted_cb);
	osip_ua_signal_connect(ua,"BYE",bye_cb);

	/* create a sdp handler and bind it to the ua */
	sdph=sdp_handler_new();
	sdp_handler_set_fcn_get_offer(SDP_HANDLER(sdph),generate_client_offer);
	osip_ua_add_body_handler(ua,sdph);
	
	
	
	
	dialog=osip_dialog_new(ua);
	osip_dialog_invite(dialog,callee,"application/sdp");

	osip_trace(OSIP_INFO3,("info: sleeping...\n"));
	while (run_cond) sleep(1);
	//scanf("%i",&err);
	osip_dialog_bye(dialog);
	sleep(2);
	osip_trace(OSIP_ERROR,("info: exiting...\n"));
	osip_ua_destroy(ua);
	exit(1);
}

/* return an port number */
char *server_get_port(SdpContext *sdpc,int a)
{
	return sgetcopy("14000");
}

void server(void)
{
	OsipUA *ua;
	BodyHandler *sdph;
	OsipDialog *dia;
	/* create the ua */
	ua=osip_ua_new();
	/* set a sending socket different to let client and server running on the same host */
	osip_manager_set_send_port(ua->manager,5070);
	osip_ua_set_contact(ua,local_sipurl);
	osip_ua_set_max_dialogs(ua,1);
	/* set some callbacks */
	osip_ua_signal_connect(ua,"INVITE",invite_cb);
	osip_ua_signal_connect(ua,"BYE",bye_cb);
	
	/* create a sdp handler to give the ua some sdp capabilities */
	sdph=sdp_handler_new();
	/* add support for some audio codecs*/
	sdp_handler_add_support_for_audio_codec(SDP_HANDLER(sdph),sgetcopy("0"),sgetcopy("RTP/AVP"),
											sgetcopy("pcmu/8000/1"));
	sdp_handler_add_support_for_audio_codec(SDP_HANDLER(sdph),sgetcopy("3"),sgetcopy("RTP/AVP"),
											sgetcopy("gsm/8000/1"));
	sdp_handler_add_support_for_audio_codec(SDP_HANDLER(sdph),sgetcopy("115"),sgetcopy("RTP/AVP"),
											sgetcopy("lpc10-15/8000/1"));
	sdp_handler_set_fcn_get_audio_port(SDP_HANDLER(sdph),server_get_port);
	osip_ua_add_body_handler(ua,sdph);
	
	osip_trace(OSIP_ERROR,("sleeping...\n"));
	while (run_cond) sleep(1);	
	/* end the pending dialog, if exist */
	dia=osip_ua_get_dialog(ua,0);
	if (dia!=NULL){
		osip_dialog_bye(dia);
	}
	sleep(1);	/* to wait for the 200 OK */
	//scanf("%i",&err);
	osip_trace(OSIP_INFO3,("exiting...\n"));
	osip_ua_destroy(ua);
	exit(1);
}

int invite_cb(OsipDialog *cl,transaction_t *trn, sip_t *sipmsg,void*p)
{
	char *from;
	
	from_2char(sipmsg->from,&from);
	osip_trace(OSIP_ERROR,("Receiving invite from %s.\n",from));
	sfree(from);
	sleep(1);
	osip_trace(OSIP_ERROR,("Accepting invite.\n"));
	osip_dialog_accept_invite(cl,trn);
	return(0);
}

int bye_cb(OsipDialog *cl,transaction_t *trn,sip_t *sipmsg,void*p)
{
	char *from;
	from_2char(sipmsg->from,&from);
	osip_trace(OSIP_ERROR,("info: Receiving bye from %s.\n",from));
	sfree(from);
	return(0);
}	

int invite_accepted_cb(OsipDialog *cl,transaction_t *trn,sip_t *msg,void*p)
{
	
	osip_trace(OSIP_ERROR,("info: Invite accepted by remote server!\n"));
	
	return(0);
}

void stop_func(int signum)
{
	run_cond=0;
}

