 /*
  * The osipua library is a library based on oSIP that implements CallLeg and User Agent
  * level.
  * Copyright (C) 2001  Simon MORLAT simon.morlat@free.fr
  * Aymeric MOIZARD jack@atosc.org
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  * 
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#include "sdphandler.h"

int
accept_all_codecs (SdpContext * ctx, char *p1, char *p2, char *p3, char *p4)
{
	/* by default all supported codecs are accepted */
	return 0;
}

int
payload_init (payload_t ** payload)
{
	*payload = (payload_t *) smalloc (sizeof (payload_t));
	if (*payload == NULL)
		return -1;
	(*payload)->payload = NULL;
	(*payload)->number_of_port = NULL;
	(*payload)->proto = NULL;
	(*payload)->c_nettype = NULL;
	(*payload)->c_addrtype = NULL;
	(*payload)->c_addr = NULL;
	(*payload)->c_addr_multicast_ttl = NULL;
	(*payload)->c_addr_multicast_int = NULL;
	(*payload)->a_rtpmap = NULL;
	return 0;
}

void
payload_free (payload_t * payload)
{
	sfree (payload->payload);
	sfree (payload->number_of_port);
	sfree (payload->proto);
	sfree (payload->c_nettype);
	sfree (payload->c_addrtype);
	sfree (payload->c_addr);
	sfree (payload->c_addr_multicast_ttl);
	sfree (payload->c_addr_multicast_int);
	sfree (payload->a_rtpmap);
}

static SdpHandlerClass *sdp_handler_class = NULL;

void
sdp_handler_init (SdpHandler * config)
{
	body_handler_init (BODY_HANDLER (config));

	config->o_username = NULL;
	config->o_session_id = NULL;
	config->o_session_version = NULL;
	config->o_nettype = NULL;
	config->o_addrtype = NULL;
	config->o_addr = NULL;

	config->c_nettype = NULL;
	config->c_addrtype = NULL;
	config->c_addr = NULL;
	config->c_addr_multicast_ttl = NULL;
	config->c_addr_multicast_int = NULL;

	/* supported codec for the SIP User Agent */
	config->audio_codec = (list_t *) smalloc (sizeof (list_t));
	list_init (config->audio_codec);
	config->video_codec = (list_t *) smalloc (sizeof (list_t));
	list_init (config->video_codec);
	config->other_codec = (list_t *) smalloc (sizeof (list_t));
	list_init (config->other_codec);

	config->fcn_set_info = NULL;
	config->fcn_set_uri = NULL;
	config->fcn_set_emails = NULL;
	config->fcn_set_phones = NULL;
	config->fcn_set_attributes = NULL;
	config->fcn_accept_audio_codec =
		(int (*)(void *, char *, char *, int, char *))
		accept_all_codecs;
	config->fcn_accept_video_codec =
		(int (*)(void *, char *, char *, int, char *))
		accept_all_codecs;
	config->fcn_accept_other_codec =
		(int (*)(void *, char *, char *, char *, char *))
		accept_all_codecs;
	config->fcn_get_audio_port = NULL;
	config->fcn_get_video_port = NULL;
	config->fcn_get_other_port = NULL;

	config->sdp_template = NULL;
}


void
sdp_handler_class_init (SdpHandlerClass * klass)
{
	BodyHandlerClass *bklass = BODY_HANDLER_CLASS (klass);
	body_handler_class_init (bklass);
	bklass->mime_type = "application/sdp";
	bklass->_body_context_new = (BodyContextNewFunc) sdp_context_new;
	bklass->_init = (BodyHandlerFunc) sdp_handler_guess_config;

}

BodyHandler *
sdp_handler_new ()
{
	SdpHandler *obj;
	if (sdp_handler_class == NULL)
	{
		sdp_handler_class = smalloc (sizeof (SdpHandlerClass));
		sdp_handler_class_init (sdp_handler_class);
	}
	obj = smalloc (sizeof (SdpHandler));
	BODY_HANDLER (obj)->klass = BODY_HANDLER_CLASS (sdp_handler_class);
	sdp_handler_init (obj);
	return BODY_HANDLER (obj);
}


void
sdp_handler_uninit (SdpHandler * config)
{
	sfree (config->o_username);
	sfree (config->o_session_id);
	sfree (config->o_session_version);
	sfree (config->o_nettype);
	sfree (config->o_addrtype);
	sfree (config->o_addr);

	sfree (config->c_nettype);
	sfree (config->c_addrtype);
	sfree (config->c_addr);
	sfree (config->c_addr_multicast_ttl);
	sfree (config->c_addr_multicast_int);

	list_special_free (config->audio_codec,
			   (void *(*)(void *)) &payload_free);
	sfree (config->audio_codec);
	list_special_free (config->video_codec,
			   (void *(*)(void *)) &payload_free);
	sfree (config->video_codec);
	list_special_free (config->other_codec,
			   (void *(*)(void *)) &payload_free);
	sfree (config->other_codec);
	if (config->sdp_template != NULL)
	{
		sdp_free (config->sdp_template);
		sfree (config->sdp_template);
		config->sdp_template = NULL;
	}
}

void
sdp_handler_destroy (SdpHandler * obj)
{
	sdp_handler_uninit (obj);
	sfree (obj);
}


int
sdp_handler_set_o_username (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->o_username = tmp;
	return 0;
}

int
sdp_handler_set_o_session_id (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->o_session_id = tmp;
	return 0;
}

int
sdp_handler_set_o_session_version (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->o_session_version = tmp;
	return 0;
}

int
sdp_handler_set_o_nettype (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->o_nettype = tmp;
	return 0;
}

int
sdp_handler_set_o_addrtype (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->o_addrtype = tmp;
	return 0;
}

int
sdp_handler_set_o_addr (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->o_addr = tmp;
	return 0;
}

int
sdp_handler_set_c_nettype (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->c_nettype = tmp;
	return 0;
}

int
sdp_handler_set_c_addrtype (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->c_addrtype = tmp;
	return 0;
}

int
sdp_handler_set_c_addr (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->c_addr = tmp;
	return 0;
}

int
sdp_handler_set_c_addr_multicast_ttl (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->c_addr_multicast_ttl = tmp;
	return 0;
}

int
sdp_handler_set_c_addr_multicast_int (SdpHandler * config, char *tmp)
{
	if (config == NULL)
		return -1;
	config->c_addr_multicast_int = tmp;
	return 0;
}

int
sdp_handler_add_support_for_audio_codec_full (SdpHandler * config,
					      char *payload,
					      char *number_of_port,
					      char *proto, char *c_nettype,
					      char *c_addrtype, char *c_addr,
					      char *c_addr_multicast_ttl,
					      char *c_addr_multicast_int,
					      char *a_rtpmap)
{
	int i;
	payload_t *my_payload;
	i = payload_init (&my_payload);
	if (i != 0)
		return -1;
	my_payload->payload = payload;
	my_payload->number_of_port = number_of_port;
	my_payload->proto = proto;
	my_payload->c_nettype = c_nettype;
	my_payload->c_addrtype = c_addrtype;
	my_payload->c_addr = c_addr;
	my_payload->c_addr_multicast_ttl = c_addr_multicast_ttl;
	my_payload->c_addr_multicast_int = c_addr_multicast_int;
	my_payload->a_rtpmap = a_rtpmap;
	list_add (config->audio_codec, my_payload, -1);
	return 0;
}

int
sdp_handler_add_support_for_audio_codec (SdpHandler * config, char *payload,
					 char *proto, char *a_rtpmap)
{
	return sdp_handler_add_support_for_audio_codec_full (config, payload,
							     sgetcopy ("1"),
							     proto,
							     sgetcopy
							     (config->
							      c_nettype),
							     sgetcopy
							     (config->
							      c_addrtype),
							     sgetcopy
							     (config->c_addr),
							     NULL, NULL,
							     a_rtpmap);
}

int
sdp_handler_add_support_for_video_codec (SdpHandler * config, char *payload,
					 char *proto, char *a_rtpmap) 
{
	return sdp_handler_add_support_for_video_codec_full (config, payload,
							     sgetcopy ("1"),
							     proto,
							     sgetcopy
							     (config->
							      c_nettype),
							     sgetcopy
							     (config->
							      c_addrtype),
							     sgetcopy
							     (config->c_addr),
							     NULL, NULL,
							     a_rtpmap);
}

int
sdp_handler_add_support_for_video_codec_full (SdpHandler * config,
					      char *payload,
					      char *number_of_port,
					      char *proto, char *c_nettype,
					      char *c_addrtype, char *c_addr,
					      char *c_addr_multicast_ttl,
					      char *c_addr_multicast_int,
					      char *a_rtpmap)
{
	int i;
	payload_t *my_payload;
	i = payload_init (&my_payload);
	if (i != 0)
		return -1;
	my_payload->payload = payload;
	my_payload->number_of_port = number_of_port;
	my_payload->proto = proto;
	my_payload->c_nettype = c_nettype;
	my_payload->c_addrtype = c_addrtype;
	my_payload->c_addr = c_addr;
	my_payload->c_addr_multicast_ttl = c_addr_multicast_ttl;
	my_payload->c_addr_multicast_int = c_addr_multicast_int;
	my_payload->a_rtpmap = a_rtpmap;
	list_add (config->video_codec, my_payload, -1);
	return 0;
}

int
sdp_handler_add_support_for_other_codec_full (SdpHandler * config,
					      char *payload,
					      char *number_of_port,
					      char *proto, char *c_nettype,
					      char *c_addrtype, char *c_addr,
					      char *c_addr_multicast_ttl,
					      char *c_addr_multicast_int,
					      char *a_rtpmap)
{
	int i;
	payload_t *my_payload;
	i = payload_init (&my_payload);
	if (i != 0)
		return -1;
	my_payload->payload = payload;
	my_payload->number_of_port = number_of_port;
	my_payload->proto = proto;
	my_payload->c_nettype = c_nettype;
	my_payload->c_addrtype = c_addrtype;
	my_payload->c_addr = c_addr;
	my_payload->c_addr_multicast_ttl = c_addr_multicast_ttl;
	my_payload->c_addr_multicast_int = c_addr_multicast_int;
	my_payload->a_rtpmap = a_rtpmap;
	list_add (config->other_codec, my_payload, -1);
	return 0;
}

int
sdp_handler_remove_audio_payloads (SdpHandler * config)
{
	list_special_free (config->audio_codec,
			   (void *(*)(void *)) &payload_free);
	return 0;
}

int
sdp_handler_remove_video_payloads (SdpHandler * config)
{
	list_special_free (config->video_codec,
			   (void *(*)(void *)) &payload_free);
	return 0;
}

int
sdp_handler_remove_other_payloads (SdpHandler * config)
{
	list_special_free (config->other_codec,
			   (void *(*)(void *)) &payload_free);
	return 0;
}

payload_t *
sdp_handler_find_audio_payload (SdpHandler * config, char *payload,
				char *rtpmap)
{
	payload_t *my;
	int length = strlen (payload);
	int pos = 0;
	int tmp;
	char *rtpdef;
	if (payload == NULL)
		return NULL;
	while (!list_eol (config->audio_codec, pos))
	{
		my = (payload_t *) list_get (config->audio_codec, pos);
		if (strlen (my->payload) == length)
			if (0 == strncmp (my->payload, payload, length))
			{
				/* simon: check the rtpmap concordance */
				if (rtpmap == NULL)
					return my;
				rtpdef = alloca (strlen (rtpmap));
				rtpdef[0] = '\0';
				sscanf (rtpmap, "%i %s", &tmp, rtpdef);
				if (strncasecmp
				    (rtpdef, my->a_rtpmap,
				     strlen (rtpdef)) == 0)
					return my;
			}
		pos++;
	}
	return NULL;
}

payload_t *
sdp_handler_find_video_payload (SdpHandler * config, char *payload)
{
	payload_t *my;
	int length = strlen (payload);
	int pos = 0;
	if (payload == NULL)
		return NULL;
	while (!list_eol (config->video_codec, pos))
	{
		my = (payload_t *) list_get (config->video_codec, pos);
		if (strlen (my->payload) == length)
			if (0 == strncmp (my->payload, payload, length))
				return my;
		pos++;
	}
	return NULL;
}

payload_t *
sdp_handler_find_other_payload (SdpHandler * config, char *payload)
{
	payload_t *my;
	int length = strlen (payload);
	int pos = 0;
	if (payload == NULL)
		return NULL;
	while (!list_eol (config->other_codec, pos))
	{
		my = (payload_t *) list_get (config->other_codec, pos);
		if (strlen (my->payload) == length)
			if (0 == strncmp (my->payload, payload, length))
				return my;
		pos++;
	}
	return NULL;
}

int
sdp_handler_set_fcn_set_info (SdpHandler * config,
			      int (*fcn) (SdpContext *, sdp_t *))
{
	if (config == NULL)
		return -1;
	config->fcn_set_info = (int (*)(void *, sdp_t *)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_set_uri (SdpHandler * config,
			     int (*fcn) (SdpContext *, sdp_t *))
{
	if (config == NULL)
		return -1;
	config->fcn_set_uri = (int (*)(void *, sdp_t *)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_set_emails (SdpHandler * config,
				int (*fcn) (SdpContext *, sdp_t *))
{
	if (config == NULL)
		return -1;
	config->fcn_set_emails = (int (*)(void *, sdp_t *)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_set_phones (SdpHandler * config,
				int (*fcn) (SdpContext *, sdp_t *))
{
	if (config == NULL)
		return -1;
	config->fcn_set_phones = (int (*)(void *, sdp_t *)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_set_attributes (SdpHandler * config,
				    int (*fcn) (SdpContext *, sdp_t *, int))
{
	if (config == NULL)
		return -1;
	config->fcn_set_attributes = (int (*)(void *, sdp_t *, int)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_accept_audio_codec (SdpHandler * config,
					int (*fcn) (SdpContext *, char *,
						    char *, int, char *))
{
	if (config == NULL)
		return -1;
	config->fcn_accept_audio_codec = (int (*)(void *, char *,
						  char *, int, char *)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_accept_video_codec (SdpHandler * config,
					int (*fcn) (SdpContext *, char *,
						    char *, int, char *))
{
	if (config == NULL)
		return -1;
	config->fcn_accept_video_codec = (int (*)(void *, char *,
						  char *, int, char *)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_accept_other_codec (SdpHandler * config,
					int (*fcn) (SdpContext *, char *,
						    char *, char *, char *))
{
	if (config == NULL)
		return -1;
	config->fcn_accept_other_codec = (int (*)(void *, char *,
						  char *, char *,
						  char *)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_get_audio_port (SdpHandler * config,
				    char *(*fcn) (SdpContext *, int))
{
	if (config == NULL)
		return -1;
	config->fcn_get_audio_port = (char *(*)(void *, int)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_get_video_port (SdpHandler * config,
				    char *(*fcn) (SdpContext *, int))
{
	if (config == NULL)
		return -1;
	config->fcn_get_video_port = (char *(*)(void *, int)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_get_other_port (SdpHandler * config,
				    char *(*fcn) (SdpContext *, int))
{
	if (config == NULL)
		return -1;
	config->fcn_get_other_port = (char *(*)(void *, int)) fcn;
	return 0;
}

int
sdp_handler_set_fcn_get_offer (SdpHandler * config,
			       sdp_t * (*fcn) (SdpContext *, sdp_t *))
{
	if (config == NULL)
		return -1;
	config->fcn_get_offer = (sdp_t * (*)(void *, sdp_t *)) fcn;
	return 0;
}

int
sdp_partial_clone (SdpHandler * config, SdpContext * con, sdp_t * remote,
		   sdp_t ** dest)
{
	int i;
	sdp_v_version_set (*dest, sgetcopy ("0"));

	/* those fields MUST be set */
	sdp_o_origin_set (*dest,
			  sgetcopy (config->o_username),
			  sgetcopy (config->o_session_id),
			  sgetcopy (config->o_session_version),
			  sgetcopy (config->o_nettype),
			  sgetcopy (config->o_addrtype),
			  sgetcopy (config->o_addr));
	sdp_s_name_set (*dest, sgetcopy (remote->s_name));
	if (config->fcn_set_info != NULL)
		config->fcn_set_info (con, *dest);
	if (config->fcn_set_uri != NULL)
		config->fcn_set_uri (con, *dest);
	if (config->fcn_set_emails != NULL)
		config->fcn_set_emails (con, *dest);
	if (config->fcn_set_phones != NULL)
		config->fcn_set_phones (con, *dest);
	if (config->c_nettype != NULL)
		sdp_c_connection_add (*dest, -1,
				      sgetcopy (config->c_nettype),
				      sgetcopy (config->c_addrtype),
				      sgetcopy (config->c_addr),
				      sgetcopy (config->c_addr_multicast_ttl),
				      sgetcopy (config->
						c_addr_multicast_int));

	{			/* offer-answer draft says we must copy the "t=" line */
		char *tmp = sdp_t_start_time_get (remote, 0);
		char *tmp2 = sdp_t_stop_time_get (remote, 0);
		if (tmp == NULL || tmp2 == NULL)
			return -1;	/* no t line?? */
		i = sdp_t_time_descr_add (*dest, sgetcopy (tmp),
					  sgetcopy (tmp2));
		if (i != 0)
			return -1;
	}
	if (config->fcn_set_attributes != NULL)
		config->fcn_set_attributes (con, *dest, -1);
	return 0;
}

int
sdp_confirm_media (SdpHandler * config, SdpContext * context, sdp_t * remote,
		   sdp_t ** dest)
{
	char *payload;
	sdp_attribute_t *attr;
	char *tmp, *tmp2, *tmp3, *tmp4, *rtpmap;
	int ret;
	int i;
	int k;
	int audio_qty = 0;	/* accepted audio line: do not accept more than one */
	int video_qty = 0;
	i = 0;
	while (!sdp_endof_media (remote, i))
	{
		tmp = sdp_m_media_get (remote, i);
		tmp2 = sdp_m_port_get (remote, i);
		tmp3 = sdp_m_number_of_port_get (remote, i);
		tmp4 = sdp_m_proto_get (remote, i);

		if (tmp == NULL)
			return -1;
		sdp_m_media_add (*dest, sgetcopy (tmp), sgetcopy ("0"),
				 NULL, sgetcopy (tmp4));
		k = 0;
		if (0 == strncmp (tmp, "audio", 5))
		{
			do
			{
				payload = sdp_m_payload_get (remote, i, k);
				if (payload != NULL)
				{
					payload_t *my_payload;
					attr = sdp_attribute_get (remote, i,
								  k);
					/* simon addon: check rtpmap concordance */
					rtpmap = NULL;
					if (attr != NULL)
					{
						if ((strcmp
						     (attr->a_att_field,
						      "rtpmap") == 0)
						    && (attr->a_att_value !=
							NULL))
							rtpmap = attr->
								a_att_value;
					}
					my_payload =
						sdp_handler_find_audio_payload
						(config, payload, rtpmap);
					if (my_payload != NULL)	/* payload is supported */
					{
						ret = -1;	/* somtimes, codec can be refused even if supported */
						if (config->
						    fcn_accept_audio_codec !=
						    NULL)
							ret = config->
								fcn_accept_audio_codec
								(context,
								 tmp2, tmp3,
								 audio_qty,
								 payload);
						if (0 == ret)
						{
							sdp_m_payload_add
								(*dest, i,
								 sgetcopy
								 (payload));
							if (my_payload->
							    a_rtpmap != NULL)
							{
								rtpmap = smalloc (strlen (my_payload->a_rtpmap) + strlen (my_payload->payload) + 2);
								sprintf (rtpmap, "%s %s", my_payload->payload, my_payload->a_rtpmap);
								/* fix by Simon: add create the complete rtpmap from the rtpmap description */
								sdp_a_attribute_add
									(*dest,
									 i,
									 sgetcopy
									 ("rtpmap"),
									 rtpmap);
							}
							if (my_payload->
							    c_nettype != NULL)
							{
								sdp_media_t
									*med =
									list_get
									((*dest)->m_medias, i);
								if (list_eol
								    (med->
								     c_connections,
								     0))
									sdp_c_connection_add
										(*dest,
										 i,
										 sgetcopy
										 (my_payload->
										  c_nettype),
										 sgetcopy
										 (my_payload->
										  c_addrtype),
										 sgetcopy
										 (my_payload->
										  c_addr),
										 sgetcopy
										 (my_payload->
										  c_addr_multicast_ttl),
										 sgetcopy
										 (my_payload->
										  c_addr_multicast_int));
							}
						}
					}
				}
				k++;
			}
			while (payload != NULL);
			if (NULL != sdp_m_payload_get (*dest, i, 0))
				audio_qty = 1;
		}
		else if (0 == strncmp (tmp, "video", 5))
		{
			do
			{
				payload = sdp_m_payload_get (remote, i, k);
				if (payload != NULL)
				{
					payload_t *my_payload =
						sdp_handler_find_video_payload
						(config, payload);
					if (my_payload != NULL)	/* payload is supported */
					{
						ret = -1;
						if (config->
						    fcn_accept_video_codec !=
						    NULL)
							ret = config->
								fcn_accept_video_codec
								(context,
								 tmp2, tmp3,
								 video_qty,
								 payload);
						if (0 == ret)
						{
							sdp_m_payload_add
								(*dest, i,
								 sgetcopy
								 (payload));
							/* TODO  set the attribute list (rtpmap..) */
							if (my_payload->
							    a_rtpmap != NULL)
								sdp_a_attribute_add
									(*dest,
									 i,
									 sgetcopy
									 ("rtpmap"),
									 sgetcopy
									 (my_payload->
									  a_rtpmap));
							if (my_payload->
							    c_nettype != NULL)
							{
								sdp_media_t
									*med =
									list_get
									((*dest)->m_medias, i);
								if (list_eol
								    (med->
								     c_connections,
								     0))
									sdp_c_connection_add
										(*dest,
										 i,
										 sgetcopy
										 (my_payload->
										  c_nettype),
										 sgetcopy
										 (my_payload->
										  c_addrtype),
										 sgetcopy
										 (my_payload->
										  c_addr),
										 sgetcopy
										 (my_payload->
										  c_addr_multicast_ttl),
										 sgetcopy
										 (my_payload->
										  c_addr_multicast_int));
							}
						}
					}
				}
				k++;
			}
			while (payload != NULL);
			if (NULL != sdp_m_payload_get (*dest, i, 0))
				video_qty = 1;
		}
		else
		{
			do
			{
				payload = sdp_m_payload_get (remote, i, k);
				if (payload != NULL)
				{
					payload_t *my_payload =
						sdp_handler_find_other_payload
						(config, payload);
					if (my_payload != NULL)	/* payload is supported */
					{
						ret = -1;
						if (config->
						    fcn_accept_other_codec !=
						    NULL)
							ret = config->
								fcn_accept_other_codec
								(context, tmp,
								 tmp2, tmp3,
								 payload);
						if (0 == ret)
						{
							sdp_m_payload_add
								(*dest, i,
								 sgetcopy
								 (payload));
							/* rtpmap has no meaning here! */
							if (my_payload->
							    c_nettype != NULL)
							{
								sdp_media_t
									*med =
									list_get
									((*dest)->m_medias, i);
								if (list_eol
								    (med->
								     c_connections,
								     0))
									sdp_c_connection_add
										(*dest,
										 i,
										 sgetcopy
										 (my_payload->
										  c_nettype),
										 sgetcopy
										 (my_payload->
										  c_addrtype),
										 sgetcopy
										 (my_payload->
										  c_addr),
										 sgetcopy
										 (my_payload->
										  c_addr_multicast_ttl),
										 sgetcopy
										 (my_payload->
										  c_addr_multicast_int));
							}
						}
					}
				}
				k++;
			}
			while (payload != NULL);
		}
		i++;
	}
	return 0;
}

int
sdp_handler_execute_negociation (SdpHandler * config, SdpContext * context)
{
	int m_lines_that_match = 0;
	sdp_t *remote;
	sdp_t *local;
	int i, k=0;
	char *rtpmap, *mapdef, *media;

	if (context == NULL)
		return -1;
	remote = context->remote;
	if (remote == NULL)
		return -1;

	i = sdp_init (&local);
	if (i != 0)
		return -1;

	if (0 != strncmp (remote->v_version, "0", -1))
	{
		sdp_free (local);
		sfree (local);
		/*      sdp_context->fcn_wrong_version(context); */
		return 406;	/* Not Acceptable */
	}

	i = sdp_partial_clone (config, context, remote, &local);
	if (i != 0)
	{
		sdp_free (local);
		sfree (local);
		return -1;
	}
	i = sdp_confirm_media (config, context, remote, &local);
	if (i != 0)
	{
		sdp_free (local);
		sfree (local);
		return i;
	}

	i = 0;
	while (!sdp_endof_media (local, i))
	{
		/* this is to refuse each line with no codec that matches! */
		if (NULL == sdp_m_payload_get (local, i, 0))
		{
			sdp_media_t *med = list_get ((local)->m_medias, i);
			char *str = sdp_m_payload_get (remote, i, 0);
			sdp_m_payload_add (local, i, sgetcopy (str));
			sfree (med->m_port);
			med->m_port = sgetcopy ("0");	/* refuse this line */
		}
		else
		{		/* number of "m" lines that match */
			sdp_media_t *med = list_get (local->m_medias, i);
			/* if there is only one payload accepted, test that it is not a telephone-event payload */
			while (sdp_m_payload_get (local, i, k) != NULL)
				k++;
			if (k == 1)
			{
				sdp_attribute_t *attr;
				int pt;
				attr = sdp_attribute_get (local, i, k);
				if (attr != NULL)
				{
					if ((strcmp
					     (attr->a_att_field,
					      "rtpmap") == 0)
					    && (attr->a_att_value != NULL))
					{
						rtpmap = attr->a_att_value;
						mapdef = alloca (strlen
								 (rtpmap));
						sscanf (rtpmap, "%i %s", &pt,
							mapdef);
						if (mapdef != NULL)
						{
							if (strncasecmp
							    (mapdef,
							     "telephone-event",
							     strlen
							     ("telephone-event")
							     == 0))
							{
								/* cannot accept this line because only telephone-event is not ok */
								m_lines_that_match--;
							}
						}
					}
				}
			}
			m_lines_that_match++;
			sfree (med->m_port);
			media = sdp_m_media_get (local, i);
			if (strncasecmp (media, "audio",strlen("audio")) == 0)
			{
				if (config->fcn_get_audio_port != NULL)
					med->m_port =config->fcn_get_audio_port (context,i);
				else
					med->m_port = sgetcopy ("0");	/* should never happen */
			}else if (strncasecmp (media, "video",strlen("video")) == 0){
				if (config->fcn_get_video_port != NULL)
					med->m_port =config->fcn_get_video_port (context,i);
				else
					med->m_port = sgetcopy ("0");	/* should never happen */
			}else{
				if (config->fcn_get_other_port != NULL)
					med->m_port =config->fcn_get_other_port (context,i);
				else
					med->m_port = sgetcopy ("0");	/* should never happen */
			}
		}


		i++;
	}
	if (m_lines_that_match > 0)
	{
		context->localneg = local;
		return 200;
	}
	else
	{
		sdp_free (local);
		sfree (local);
		return 415;
	}

}


/* guess the sdp config from the ua config, and generate a template sdp */
void
sdp_handler_guess_config (SdpHandler * obj)
{
	OsipUA *ua = BODY_HANDLER (obj)->ua;
	url_t *url = contact_geturl (ua->contact);
	sdp_t *local;
	int err;

	if (ua == NULL)
	{
		osip_trace (OSIP_ERROR, ("SdpHandler: ua is NULL"));
		return;
	}

	sdp_init (&local);

	sdp_v_version_set (local, sgetcopy ("0"));
	sdp_o_origin_set (local, sgetcopy (url_getusername (url)),
			  sgetcopy ("123456"), sgetcopy ("654321"),
			  sgetcopy ("IN"), sgetcopy ("IP4"),
			  sgetcopy (ua->ua_ip4addr));
	sdp_s_name_set (local, sgetcopy ("A conversation"));
	sdp_c_connection_add (local, -1,
			      sgetcopy ("IN"), sgetcopy ("IP4"),
			      sgetcopy (ua->ua_ip4addr), NULL, NULL);
	sdp_t_time_descr_add (local, sgetcopy ("0"), sgetcopy ("0"));

	sdp_handler_set_o_username (obj, sgetcopy (url_getusername (url)));
	sdp_handler_set_o_session_id (obj, sgetcopy ("123456"));
	sdp_handler_set_o_session_version (obj, sgetcopy ("654321"));
	sdp_handler_set_o_nettype (obj, sgetcopy ("IN"));
	sdp_handler_set_o_addrtype (obj, sgetcopy ("IP4"));
	sdp_handler_set_o_addr (obj, sgetcopy (ua->ua_ip4addr));


	sdp_handler_set_c_nettype (obj, sgetcopy ("IN"));
	sdp_handler_set_c_addrtype (obj, sgetcopy ("IP4"));
	sdp_handler_set_c_addr (obj, sgetcopy (ua->ua_ip4addr));
	if (obj->sdp_template != NULL)
	{
		sdp_free (obj->sdp_template);
		sfree (obj->sdp_template);
	}
	obj->sdp_template = local;
}
