/*
  The osipua library is a library based on oSIP that implements CallLeg and User Agent
  level.
  Copyright (C) 2001  Simon MORLAT simon.morlat@free.fr
  											Aymeric MOIZARD jack@atosc.org
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include "udp.h"
#include "osipmanager.h"
#include "osipua.h"

extern OsipManager *def_manager;

static int udp_receive (OsipManager * manager);

void osipua_distribute_event(OsipManager *m, sipevent_t *ev)
{
	osip_t *config=m->config;
	transaction_t *trn;
	int err;
	err=osip_find_transaction_and_add_event(config,ev);
	if (err<0){
		/* check if it is not an ack for invite (not handled by osip)*/
		if (ev->type==RCV_REQACK){
			/* we will do something here later FIXME*/
			sfree(ev->sip);
			sfree(ev);
		}
		else if (ev->type==RCV_STATUS_2XX){
			/* this a retransmission of 200 Ok */
			sfree(ev->sip);
			sfree(ev);
		}
		else if (EVT_IS_INCOMINGREQ(ev)){
			/* not an ack, of course */
			trn=osip_create_transaction(config,ev);
		}
	}else{
		/* an existing transaction was found */
		
	}
}


void *
sipd_thread (void *managerp)
{
	long int time_elapsed;
	struct timeval t1, t2;
	OsipManager *manager = (OsipManager *) managerp;
	sipevent_t *sipevent;
	int err;
	transaction_t *trn;
	/* SIP_MESSAGE_MAX_LENGTH should be set to 66001
	 * to improve interopperability */
	/* the main loop */
	osip_trace (OSIP_INFO1, ("Entering osipua thread.\n"));
	manager->thread_pid=getpid();
	gettimeofday (&t1, NULL);
	while (manager->udp_run_cond)
	{
		err=udp_receive (manager);
		/* lock the access to osip stack */
		smutex_lock(manager->mutex);
		gettimeofday (&t2, NULL);

		time_elapsed =
			((t2.tv_sec - t1.tv_sec) * 1000) +
			((t2.tv_usec - t1.tv_usec) / 1000);

		/* execute timers if needed */
		//printf("Time elapsed: %i\n", time_elapsed);
		if (time_elapsed > 500)	/*ms */
		{
			//printf("Executing timers\n");
			t1.tv_sec = t2.tv_sec;
			t1.tv_usec = t2.tv_usec;
			osip_timers_ict_execute (manager->config);
			osip_timers_ist_execute (manager->config);
			osip_timers_nict_execute (manager->config);
			osip_timers_nist_execute (manager->config);
		}


		//printf("Executing osip stack..\n");
		osip_ict_execute (manager->config);
		osip_nict_execute (manager->config);
		osip_ist_execute (manager->config);
		osip_nist_execute (manager->config);
		
		/* free killed transaction */
		while ( (trn=fifo_tryget(&manager->garbage_trn))!=NULL){
			transaction_free(trn);
			sfree(trn);
		}
		
		/* unlock osip*/
		smutex_unlock(manager->mutex);
	}
	osip_trace (OSIP_INFO1, ("Exiting osipua thread.\n"));
	return NULL;
}

/* this function waits 100 ms until a packet arrives , return 1 if a packet is received, 0 else */
static int
udp_receive (OsipManager * manager)
{
	int i, j, k, nb;
	fd_set osipfdset;
	struct timeval timeout;

	memcpy (&osipfdset, &manager->udpfdset, sizeof (fd_set));
	timeout.tv_sec = manager->recv_tout.tv_sec;	/* timeout can be modified by linux !! */
	timeout.tv_usec = manager->recv_tout.tv_usec;
	nb = select (manager->max_udpfd + 1, &osipfdset, NULL, NULL,
		     &timeout);
	/* if something to read on the control file descriptor, then exit thread */
	if (nb > 0)
	{
		if (FD_ISSET (manager->udp_unblock_fd, &osipfdset))
		{
			//printf("Receiving something on unblock_fd\n");
			read (manager->udp_unblock_fd, manager->udp_buf,
			      SIP_MESSAGE_MAX_LENGTH);
		}
		//printf("Something to read !!!!!!!!!!!!!!!!!\n");
		for (j = 0, k = 0; (j < OSIP_MAX_UDP_PORTS) && (k < nb); j++)
		{
			if (FD_ISSET (manager->udpfds[j], &osipfdset))
			{
				k++;
				i = recv (manager->udpfds[j],
					  manager->udp_buf,
					  SIP_MESSAGE_MAX_LENGTH, 0);
				if (i > 0)
				{
					/* Message might not end with a "\0" but we know the number of */
					/* char received! */
					sipevent_t *sipevent;
					manager->udp_buf[i] = '\0';
					osip_trace (OSIP_INFO1,
						    ("info: RECEIVING UDP MESSAGE:\n%s\n",
						     manager->udp_buf));
					smutex_lock(manager->mutex);
					sipevent = osip_parse (manager->udp_buf);
					if (sipevent != NULL)
						osipua_distribute_event(manager,sipevent);
					smutex_unlock(manager->mutex);
					
				}
				else	/* if i>0 */
				{
				  osip_trace(OSIP_ERROR,
					     ("UDP listener failed while receiving data!\n"));
				}
			}
		}
		return 1;	/* messages have been received */
	}

	return 0;		/* no messages : timeout  */
}

int
udp_send (transaction_t * trn, sip_t * sipmsg, char *host, int port, int sock)
{
	struct sockaddr_in addr;
	unsigned long int one_inet_addr;
	char *message;
	msg_2char(sipmsg,&message);
	
	osip_trace(OSIP_INFO1,("Sending message:\n%s",message));

	if ((int) (one_inet_addr = inet_addr (host)) == -1)
	{
		osip_trace (OSIP_INFO1,
			    ("error: destination is not an ip address:%s\n",
			     host));
		return -1;
	}
	else
	{
		addr.sin_addr.s_addr = one_inet_addr;
	}

	addr.sin_port = htons ((short) port);
	addr.sin_family = AF_INET;
	if (sock<=0){
		osip_trace(OSIP_INFO1,("warning: using default manager socket to send a message.\n"));
		sock=def_manager->send_sock;
	}

	if (0 > sendto (sock, (const void *) message, strlen (message), 0,
			(struct sockaddr *) &addr, sizeof (addr)))
	{
		osip_trace(OSIP_ERROR,("udp_send: could not send message using socket %i: %s\n",sock,strerror(errno)));
	}

	
	return 0;
}
